-- define AceAddon
XengaInfo = LibStub("AceAddon-3.0"):NewAddon("XengaInfo", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
if not XengaInfo then
	return
end

-- Ace config libs
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

-- get localization
local L = LibStub("AceLocale-3.0"):GetLocale("XengaInfo")

-- colors
local c_red = "|cffff0000";
local c_green = "|cff00ff00";
local c_blue = "|cff0000ff";
local c_r = "|r";

-- Debugging levels for printing messages
local d_warn = 1
local d_info = 2
local d_notice = 3
local d_off = 4
local debug_levels = {
	{
		name = L["DEBUG_NAME_WARNING"],
		color = "ff0000",
	},
	{
		name = L["DEBUG_NAME_INFORMATION"],
		color = "0066cc",
	},
	{
		name = L["DEBUG_NAME_NOTICE"],
		color = "ff8c00",
	},
	{
		name = L["DEBUG_NAME_OFF"],
		color = "",
	},
}


-- AceConfig option defaults
local defaults = {
	global = {
		enabled = true,
		debug_level = d_off,
		display = {
			clamped = true,
			locked = false,
			module_spacing = 2,
			scale = 1,
		},
		modules = {
			exp = {
				show = true,
			},
			currency_gold = {
				text_mode = false,
				show = true,
			},
			currency_gr = {
				show = true,
			},
			currency_ohr = {
				show = true,
			},
			latency = {
				show = true,
			},
			murder_power = {
				show = true,
			},
			time = {
				show = true,
			},
			zone = {
				show = true,
			},
		},
		module_order = {
			"murder_power",
			"zone",
			"exp",
			"currency_gold",
			"currency_gr",
			"currency_ohr",
			"latency",
			"time"
		}
	},
}

-- currency data that doesn't need to be stored in SavedVariables
XengaInfo.currencies = {
	gr = { -- garrison resources
		id = 824,
		name = "",
		texture = "",
	},
	ohr = { -- order hall resources
		id = 1220,
		name = "",
		texture = ""
	},
}

-- module data that doesn't need to be stored in SavedVariables
XengaInfo.modules = {
	exp = {
		frame = "XengaInfoExpFrame",
		template = "XengaInfoExpFrameTemplate",
	},
	currency_gold = {
		frame = "XengaInfoCurrencyGoldFrame",
		template = "XengaInfoCurrencyFrameTemplate",
	},
	currency_gr = {
		frame = "XengaInfoCurrencyGrFrame",
		template = "XengaInfoCurrencyFrameTemplate",
	},
	currency_ohr = {
		frame = "XengaInfoCurrencyOhrFrame",
		template = "XengaInfoCurrencyFrameTemplate",
	},
	latency = {
		frame = "XengaInfoLatencyFrame",
		template = "XengaInfoLatencyFrameTemplate",
	},
	murder_power = {
		frame = "XengaInfoMurderPowerFrame",
		template = "XengaInfoMurderPowerFrameTemplate",
	},
	time = {
		frame = "XengaInfoTimeFrame",
		template = "XengaInfoTimeFrameTemplate",
	},
	zone = {
		frame = "XengaInfoZoneFrame",
		template = "XengaInfoZoneFrameTemplate",
	},
}

-- session data that doesn't need to be stored in SavedVariables
XengaInfo.sdb = {
	char = {
		class = "",
		level = 0,
		max_level = 0,
		race = "",
	},
	currency = {
		gold = {
			last = 0,
			start = 0,
		},
		gr = { -- garrison resources
			last = 0,
			start = 0,
		},
		ohr = { -- order hall resources
			last = 0,
			start = 0,
		}
	},
	time_start = 0,
	disabled_modules = {}
}

-- AceConfig options table
XengaInfo.options = {
	cmd = {
		type = "group",
		name = "XengaInfo",
		args = {
			config = {
				type  = "execute",
				name  = L["OPT_CMD_CONFIG_NAME"],
				desc  = L["OPT_CMD_CONFIG_DESC"],
				order = 0,
				func  = function()
					AceConfigDialog:Open("XengaInfo_Core");
				end,
			},
			enable = {
				type  = "toggle",
				name  = L["OPT_CMD_ENABLE_NAME"],
				desc  = L["OPT_CMD_ENABLE_DESC"],
				order = 1,
				get   = function()
					return XengaInfo.db.global.enabled;
				end,
				set   = function(i, v)
					XengaInfo.db.global.enabled = v;
					if v then
						XengaInfo:Enable();
					else
						XengaInfo:Disable();
					end
				end,
			},
			lock = {
				type  = "toggle",
				name  = L["OPT_CMD_LOCK_NAME"],
				desc  = L["OPT_CMD_LOCK_DESC"],
				order = 2,
				get   = function()
					return XengaInfo.db.global.display.locked;
				end,
				set   = function(i, v)
					XengaInfo.db.global.display.locked = v;
				end,
			},
			reset = {
				type  = "execute",
				name  = L["OPT_CMD_RESET_NAME"],
				desc  = L["OPT_CMD_RESET_DESC"],
				order = 2,
				func  = function()
					XengaInfo:Reset();
				end,
			},
			debug = {
				type   = "select",
				name   = L["OPT_CMD_DEBUG_NAME"],
				desc   = L["OPT_CMD_DEBUG_DESC"],
				order  = 3,
				values = {"warn", "info", "notice", "off"},
				get    = function()
					return debug_levels[XengaInfo:GetDebugLevel() or 4]
				end,
				set    = function(info, v)
					if (v == 1) then 
						XengaInfo:SetDebugLevel(nil)
						XengaInfo.db.global.debug_level = nil
					else 
						XengaInfo:SetDebugLevel(v)
						XengaInfo.db.global.debug_level = v
					end
				end,
			},
		},
	},
	core = {
		type = "group",
		name = "XengaInfo",
		args = {
			-- separator
			header_general = {
				type  = "header",
				name  = L["OPT_CORE_HEADER_GENERAL_NAME"],
				order	= 0,
			},
			-- Enabled
			enabled = {
				type  = "toggle",
				name  = L["OPT_CORE_ENABLED_NAME"],
				desc  = L["OPT_CORE_ENABLED_DESC"],
				order = 1,
				get   = function()
					return XengaInfo.db.global.enabled;
				end,
				set   = function(i, v)
					XengaInfo.db.global.enabled = v;
					if v then
						XengaInfo:Enable();
					else
						XengaInfo:Disable();
					end
				end,
			},
			-- Debug Level
			debug = {
				type   = "select",
				name   = L["OPT_CORE_DEBUG_NAME"],
				desc   = L["OPT_CORE_DEBUG_DESC"],
				order  = 2,
				values = {
					L["OPT_CORE_DEBUG_SELECT_WARN"],
					L["OPT_CORE_DEBUG_SELECT_INFO"],
					L["OPT_CORE_DEBUG_SELECT_NOTICE"],
					L["OPT_CORE_DEBUG_SELECT_OFF"]
				},
				get    = function()
					return XengaInfo:GetDebugLevel() or 4;
				end,
				set    = function(info, v)
					XengaInfo:SetDebugLevel(v);
					XengaInfo.db.global.debug_level = v;
				end,
			},
			reset = {
				type  = "execute",
				width = "half",
				name  = L["OPT_CORE_RESET_NAME"],
				desc  = L["OPT_CORE_RESET_DESC"],
				order = 3,
				func  = function()
					XengaInfo:Reset();
				end,
			},
			-- separator
			header_modules = {
				type  = "header",
				name  = L["OPT_CORE_HEADER_MODULES_NAME"],
				order = 10,
			},
			-- Gold
			module_currency_gold = {
				type  = "toggle",
				width = "double",
				name  = L["OPT_CORE_MODULE_CURRENCY_GOLD_NAME"],
				desc  = L["OPT_CORE_MODULE_CURRENCY_GOLD_DESC"],
				order = 11,
				get   = function ()
					return XengaInfo.db.global.modules.currency_gold.show;
				end,
				set   = function (i, v)
					XengaInfo.db.global.modules.currency_gold.show = v;
					if v then
						XengaInfo:UpdateCurrencyText();
					end
					XengaInfo:SetFramePositions();
				end,
			},
			-- Gold text mode
			module_currency_gold_text = {
				type     = "toggle",
				width    = "double",
				name     = L["OPT_CORE_MODULE_CURRENCY_GOLD_TEXT_NAME"],
				desc     = L["OPT_CORE_MODULE_CURRENCY_GOLD_TEXT_DESC"],
				disabled = function()
					return not XengaInfo.db.global.modules.currency_gold.show;
				end,
				order    = 13,
				get      = function ()
					return XengaInfo.db.global.modules.currency_gold.text_mode;
				end,
				set      = function (i, v)
					XengaInfo.db.global.modules.currency_gold.text_mode = v;
					XengaInfo:UpdateCurrencyText();
					XengaInfo:SetFramePositions();
				end,
			},
			-- Garrison Resources
			module_currency_gr = {
				type  = "toggle",
				width = "double",
				name  = L["OPT_CORE_MODULE_CURRENCY_GR_NAME"],
				desc  = L["OPT_CORE_MODULE_CURRENCY_GR_DESC"],
				order = 15,
				get   = function ()
					return XengaInfo.db.global.modules.currency_gr.show;
				end,
				set   = function (i, v)
					XengaInfo.db.global.modules.currency_gr.show = v;
					if v then
						XengaInfo:UpdateCurrencyText();
					end
					XengaInfo:SetFramePositions();
				end,
			},
			-- Order Hall Resources
			module_currency_ohr = {
				type  = "toggle",
				width = "double",
				name  = L["OPT_CORE_MODULE_CURRENCY_OHR_NAME"],
				desc  = L["OPT_CORE_MODULE_CURRENCY_OHR_DESC"],
				order = 17,
				get   = function ()
					return XengaInfo.db.global.modules.currency_ohr.show;
				end,
				set   = function (i, v)
					XengaInfo.db.global.modules.currency_ohr.show = v;
					if v then
						XengaInfo:UpdateCurrencyText();
					end
					XengaInfo:SetFramePositions();
				end,
			},
			-- Player Experience
			module_exp = {
				type  = "toggle",
				name  = L["OPT_CORE_MODULE_EXP_NAME"],
				desc  = L["OPT_CORE_MODULE_EXP_DESC"],
				order = 14,
				get   = function ()
					return XengaInfo.db.global.modules.exp.show;
				end,
				set   = function (i, v)
					XengaInfo.db.global.modules.exp.show = v;
					if v then
						XengaInfo:UpdateExpText();
					end
					XengaInfo:SetFramePositions();
				end,
			},
			-- Latency
			module_latency = {
				type  = "toggle",
				name  = L["OPT_CORE_MODULE_LATENCY_NAME"],
				desc  = L["OPT_CORE_MODULE_LATENCY_DESC"],
				order = 19,
				get   = function ()
					return XengaInfo.db.global.modules.latency.show;
				end,
				set   = function (i, v)
					XengaInfo.db.global.modules.latency.show = v;
					if XengaInfo:ShouldScheduleLatencyTimeTimer() then
						XengaInfo:ScheduleLatencyTimeTimer();
						XengaInfo:UpdateLatencyText();
					elseif XengaInfo:ShouldStopLatencyTimeTimer() then
						XengaInfo:StopLatencyTimeTimer();
					end
					XengaInfo:SetFramePositions();
				end,
			},
			-- Murder Power
			module_murder_power = {
				type  = "toggle",
				name  = L["OPT_CORE_MODULE_MURDER_POWER_NAME"],
				desc  = L["OPT_CORE_MODULE_MURDER_POWER_DESC"],
				order = 12,
				get   = function ()
					return XengaInfo.db.global.modules.murder_power.show;
				end,
				set   = function (i, v)
					XengaInfo.db.global.modules.murder_power.show = v;
					if v then
						XengaInfo:UpdateMurderPowerText();
					end
					XengaInfo:SetFramePositions();
				end,
			},
			-- Session Time
			module_time = {
				type  = "toggle",
				name  = L["OPT_CORE_MODULE_TIME_NAME"],
				desc  = L["OPT_CORE_MODULE_TIME_DESC"],
				order = 16,
				get   = function ()
					return XengaInfo.db.global.modules.time.show;
				end,
				set   = function (i, v)
					XengaInfo.db.global.modules.time.show = v;
					if XengaInfo:ShouldScheduleLatencyTimeTimer() then
						XengaInfo:ScheduleLatencyTimeTimer();
						XengaInfo:UpdateTimeText();
					elseif XengaInfo:ShouldStopLatencyTimeTimer() then
						XengaInfo:StopLatencyTimeTimer();
					end
					XengaInfo:SetFramePositions();
				end,
			},
			-- Zone Information
			module_zone = {
				type  = "toggle",
				name  = L["OPT_CORE_MODULE_ZONE_NAME"],
				desc  = L["OPT_CORE_MODULE_ZONE_DESC"],
				order = 18,
				get   = function ()
					return XengaInfo.db.global.modules.zone.show;
				end,
				set   = function (i, v)
					XengaInfo.db.global.modules.zone.show = v;
					if v then
						XengaInfo:UpdateZoneText();
					end
					XengaInfo:SetFramePositions();
				end,
			},
			-- separator
			header_display = {
				type  = "header",
				name  = L["OPT_CORE_HEADER_DISPLAY_NAME"],
				order = 20,
			},
			-- Clamped To Screen
			frame_clamped = {
				type  = "toggle",
				name  = L["OPT_CORE_FRAME_CLAMPED_NAME"],
				desc  = L["OPT_CORE_FRAME_CLAMPED_DESC"],
				order = 21,
				get   = function ()
					return XengaInfo.db.global.display.clamped;
				end,
				set   = function (i, v)
					XengaInfo.db.global.display.clamped = v;
					XengaInfoFrame:SetClampedToScreen(v);
				end,
			},
			-- Frame Scale
			frame_scale = {
				type  = "range",
				width = "double",
				name  = L["OPT_CORE_FRAME_SCALE_NAME"],
				desc  = L["OPT_CORE_FRAME_SCALE_DESC"],
				order = 22,
				min   = 0.7,
				max   = 2.0,
				step  = 0.1,
				get   = function ()
					return XengaInfo.db.global.display.scale;
				end,
				set   = function (i, v)
					XengaInfo.db.global.display.scale = v;
					XengaInfo:SetFrameScale(v);
				end,
			},
			-- Position Locked
			position_locked = {
				type  = "toggle",
				name  = L["OPT_CORE_POSITION_LOCKED_NAME"],
				desc  = L["OPT_CORE_POSITION_LOCKED_DESC"],
				order = 23,
				get   = function()
					return XengaInfo.db.global.display.locked;
				end,
				set   = function(i, v)
					XengaInfo.db.global.display.locked = v;
				end,
			},
			-- Module Spacing
			module_spacing = {
				type  = "range",
				width = "double",
				name  = L["OPT_CORE_MODULE_SPACING_NAME"],
				desc  = L["OPT_CORE_MODULE_SPACING_DESC"],
				order = 24,
				min   = 0,
				max   = 10,
				step  = 1,
				get   = function ()
					return XengaInfo.db.global.display.module_spacing;
				end,
				set   = function (i, v)
					XengaInfo.db.global.display.module_spacing = v;
					XengaInfo:SetFramePositions(v);
				end,
			},
		}
	}
}

----------------------------------------
-- initilization functions
----------------------------------------

--[[
-- Sets core data when the addon is initilized
]]
function XengaInfo:OnInitialize()
	-- get saved variables
	self.db = LibStub("AceDB-3.0"):New("XengaInfoDB", defaults, true);

	-- set initial enabled state
	self.enabledState = self.db.global.enabled;
	
	-- set up chat commands
	AceConfig:RegisterOptionsTable("XengaInfo", self.options.cmd, {"xi", "xengainfo"})
	
	-- setup options window
	AceConfig:RegisterOptionsTable("XengaInfo_Core", self.options.core);
	self.options_frame = AceConfigDialog:AddToBlizOptions("XengaInfo_Core", "XengaInfo")
	
	self:Debug(d_notice, "XengaInfo has been initialized.");
end

--[[
-- Sets up handles when the addon is enabled
]]
function XengaInfo:OnEnable()
	-- set initial char info
	self.sdb.char.level = UnitLevel('player');
	self.sdb.char.max_level = GetMaxPlayerLevel();
	self.sdb.time_start = time();
	
	-- set initial currency values
	self.sdb.currency.gold.start = GetMoney();
	self.sdb.currency.gold.last = self.sdb.currency.gold.start;
	self.sdb.currency.gr.name, self.sdb.currency.gr.start, self.sdb.currency.gr.texture = GetCurrencyInfo(self.currencies.gr.id);
	self.sdb.currency.gr.last = self.sdb.currency.gr.start;
	self.sdb.currency.ohr.name, self.sdb.currency.ohr.start, self.sdb.currency.ohr.texture = GetCurrencyInfo(self.currencies.ohr.id);
	self.sdb.currency.ohr.last = self.sdb.currency.ohr.start;
	
	-- initialize frame
	self:InitializeFrame()
	
	-- register main events
	--- character text
	self:RegisterEvent("PLAYER_LEVEL_UP", function(event, level) XengaInfo:ProcessLevelUp(level) end);
	--- murder power text
	self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", "UpdateMurderPowerText");
	self:RegisterEvent("PLAYER_PVP_KILLS_CHANGED", "UpdateMurderPowerText");
	--- zone text
	self:RegisterEvent("ZONE_CHANGED", "UpdateZoneText");
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "UpdateZoneText");
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "UpdateZoneText");
	--- exp text
	self:RegisterEvent("PLAYER_XP_UPDATE", "UpdateExpText");
	self:RegisterEvent("UPDATE_EXHAUSTION", "UpdateExpText");
	--- currency
	self:RegisterEvent("CURRENCY_DISPLAY_UPDATE", "UpdateCurrencyText");
	self:RegisterEvent("PLAYER_MONEY", "UpdateCurrencyText");
	
	-- start timers
	if self:ShouldScheduleLatencyTimeTimer() then
		self:ScheduleLatencyTimeTimer();
	end
	
	self:Printf(L["MSG_ADDON_ENABLED"], "ff00ff00");
end

--[[
-- Cleans up handles when the addon is disabled
]]
function XengaInfo:OnDisable()
	XengaInfoFrame:Hide();
	
	self:UnregisterEvent("PLAYER_LEVEL_UP");
	self:UnregisterEvent("PLAYER_EQUIPMENT_CHANGED");
	self:UnregisterEvent("PLAYER_PVP_KILLS_CHANGED");
	self:UnregisterEvent("ZONE_CHANGED");
	self:UnregisterEvent("ZONE_CHANGED_INDOORS");
	self:UnregisterEvent("ZONE_CHANGED_NEW_AREA");
	self:UnregisterEvent("PLAYER_XP_UPDATE");
	self:UnregisterEvent("UPDATE_EXHAUSTION");
	self:UnregisterEvent("CURRENCY_DISPLAY_UPDATE");
	self:UnregisterEvent("PLAYER_MONEY");
	
	self:StopLatencyTimeTimer();
	
	self:Printf(L["MSG_ADDON_DISABLED"], "ffff0000");
end

--[[
-- Initializes the frame
]]
function XengaInfo:InitializeFrame()
	-- show main frame
	XengaInfoFrame:Show();
	
	-- set frame clamp
	XengaInfoFrame:SetClampedToScreen(self.db.global.display.clamped);
	
	-- set frame scale
	XengaInfo:SetFrameScale(self.db.global.display.scale);
	
	-- configure module frames
	self:InitializeModuleFrames();
	
	-- disable level based modules
	--- experience meter
	if self.sdb.char.level == self.sdb.char.max_level then
		self:Debug(d_notice, c_green .. "XI INIT: " .. c_r .. "disabling exp module, character at max obtainable level");
		self:DisableModule("exp");
	end
	--- garrison resources
	if self.sdb.char.level < 90 then
		self:Debug(d_notice, c_green .. "XI INIT: " .. c_r .. "disabling garrison resource module, character not high enough level");
		self:DisableModule("currency_gr");
	end
	--- order hall resources
	if self.sdb.char.level < 100 then
		self:Debug(d_notice, c_green .. "XI INIT: " .. c_r .. "disabling order hall resource module, character not high enough level");
		self:DisableModule("currency_ohr");
	end
	
	-- set initial frame positions
	self:SetFramePositions();
	
	-- set initial data
	self:UpdateAllText();
end

--[[
-- Creates the initial frames and sets some graphics based elements
]]
function XengaInfo:InitializeModuleFrames()
	-- create frames
	self:CreateModuleFrame(self.modules["currency_gold"].frame, self.modules["currency_gold"].template);
	self:CreateModuleFrame(self.modules["currency_gr"].frame, self.modules["currency_gr"].template);
	self:CreateModuleFrame(self.modules["currency_ohr"].frame, self.modules["currency_ohr"].template);
	self:CreateModuleFrame(self.modules["exp"].frame, self.modules["exp"].template);
	self:CreateModuleFrame(self.modules["latency"].frame, self.modules["latency"].template);
	self:CreateModuleFrame(self.modules["murder_power"].frame, self.modules["murder_power"].template);
	self:CreateModuleFrame(self.modules["time"].frame, self.modules["time"].template);
	self:CreateModuleFrame(self.modules["zone"].frame, self.modules["zone"].template);
	
	-- set frame label text
	XengaInfoMurderPowerFrame.HKLabel:SetText(L["LABEL_HK"] .. ":");
	XengaInfoMurderPowerFrame.ItemLevelLabel:SetText(L["LABEL_ILVL"] .. ":");
	XengaInfoExpFrame.TnlLabel:SetText(L["LABEL_TNL"] .. ":");
	XengaInfoExpFrame.RestLabel:SetText(L["LABEL_EXP"] .. ":");
	XengaInfoCurrencyGoldFrame.CurrentLabel:SetText(L["LABEL_CURRENT"] .. ":");
	XengaInfoCurrencyGoldFrame.SessionLabel:SetText(L["LABEL_SESSION"] .. ":");
	XengaInfoCurrencyGrFrame.CurrentLabel:SetText(L["LABEL_CURRENT"] .. " |T" .. self.sdb.currency.gr.texture .. ":0|t :");
	XengaInfoCurrencyGrFrame.SessionLabel:SetText(L["LABEL_SESSION"] .. " |T" .. self.sdb.currency.gr.texture .. ":0|t :");
	XengaInfoCurrencyOhrFrame.CurrentLabel:SetText(L["LABEL_CURRENT"] .. " |T" .. self.sdb.currency.ohr.texture .. ":0|t :");
	XengaInfoCurrencyOhrFrame.SessionLabel:SetText(L["LABEL_SESSION"] .. " |T" .. self.sdb.currency.ohr.texture .. ":0|t :");
	XengaInfoLatencyFrame.LatencyLabel:SetText(L["LABEL_LATENCY"] .. ":");
	XengaInfoTimeFrame.TimeLabel:SetText(L["LABEL_SESSION_TIME"] .. ":");
	
	-- set frame label widths
	local label_padding = 2;
	XengaInfoMurderPowerFrame.HKLabel:SetWidth(XengaInfoMurderPowerFrame.HKLabel:GetStringWidth() + label_padding);
	XengaInfoMurderPowerFrame.ItemLevelLabel:SetWidth(XengaInfoMurderPowerFrame.ItemLevelLabel:GetStringWidth() + label_padding);
	self:NormalizeStringWidths(XengaInfoExpFrame.TnlLabel, XengaInfoExpFrame.RestLabel, label_padding);
	self:NormalizeStringWidths(XengaInfoCurrencyGoldFrame.CurrentLabel, XengaInfoCurrencyGoldFrame.SessionLabel, label_padding);
	self:NormalizeStringWidths(XengaInfoCurrencyGrFrame.CurrentLabel, XengaInfoCurrencyGrFrame.SessionLabel, label_padding);
	self:NormalizeStringWidths(XengaInfoCurrencyOhrFrame.CurrentLabel, XengaInfoCurrencyOhrFrame.SessionLabel, label_padding);
	XengaInfoLatencyFrame.LatencyLabel:SetWidth(XengaInfoLatencyFrame.LatencyLabel:GetStringWidth() + label_padding);
	XengaInfoTimeFrame.TimeLabel:SetWidth(XengaInfoTimeFrame.TimeLabel:GetStringWidth() + label_padding);
end

--[[
-- Resets the settings back to default
]]
function XengaInfo:Reset()
	self.db:ResetDB();
	
	-- make sure modules that were skipped during initial load are setup
	self:InitializeModuleFrames()
	
	-- reposition frames
	self:SetFramePositions();
	
	-- update all applicable text
	self:UpdateAllText();
	
	-- reset timers
	if self:ShouldScheduleLatencyTimeTimer() then
		self:ScheduleLatencyTimeTimer();
	end
	
	-- notify user
	self:Print(L["MSG_OPTIONS_RESET"]);
end

----------------------------------------
-- END initilization functions
----------------------------------------

----------------------------------------
-- frame display functions
----------------------------------------

--[[
-- Returns whether the frame can be moved or not
]]
function XengaInfo:CanMove()
	return not self.db.global.display.locked;
end

--[[
-- Listens for the right click event to open the settings button
]]
function XengaInfo:HandleRightClick(frame, button)
	-- open config?
	if frame:GetName() == "XengaInfoTitleFrame" then
		if IsShiftKeyDown() then
			AceConfigDialog:Open("XengaInfo_Core");
		end
	end
end

--[[
-- Creates a module frame. Since you can't actually delete a frame once it's created, this will re-use a frame that had the same name previously.
]]
function XengaInfo:CreateModuleFrame(frame_name, template)
	local parent_frame = _G["XengaInfoFrame"];
	-- make sure it doesn't already exist
	local module_frame = _G[frame_name];
	if not module_frame then
		module_frame = CreateFrame("Frame", frame_name, parent_frame, template);
		module_frame.parent = parent_frame;
	end
		
	-- setup module frame
	module_frame:Show();
	module_frame:ClearAllPoints();
	
	return module_frame;
end

--[[
-- Removes a frame from display. Since you can't actually "delete" a frame, this just hides it and clears all points so it can be reused later.
]]
function XengaInfo:RemoveModuleFrame(name)
	frame = _G[name];
	if frame then
		frame:Hide();
		--frame:ClearAllPoints();
	end
end

--[[
-- Sets the XML points for each module frame
]]
function XengaInfo:SetFramePositions()
	-- only do stuff if the main frame is visible
	if XengaInfoFrame:IsVisible() then
		local relative_frame = "XengaInfoTopBufferFrame";
		local height = _G[relative_frame]:GetHeight() + self.db.global.display.module_spacing + 6;
				
		for k,v in pairs(self.db.global.module_order) do
			-- module has an entry?
			if self.modules[v] then
				-- get frame
				local module_frame = _G[self.modules[v].frame];
					
				-- module active?
				if self:IsModuleActive(v) then			
					-- if somehow we don't have it created, create it
					if not module_frame then
						module_frame = self:CreateModuleFrame(self.modules[v].name, self.modules[v].template);
					end
					
					-- if it isn't visible, show it
					if not module_frame:IsVisible() then
						module_frame:Show();
					end
					
					-- update total height
					height = height + module_frame:GetHeight() + self.db.global.display.module_spacing;
					
					-- set position
					module_frame:SetPoint("TOP", relative_frame, "BOTTOM", 0, -self.db.global.display.module_spacing);
					
					relative_frame = module_frame;
				else
					-- make sure it's hidden
					if module_frame then
						if module_frame:IsVisible() then
							module_frame:Hide();
						end
					end
					self:Debug(d_notice, "Skipping module " .. c_green .. v .. c_r .. " " .. c_red .. "not active" .. c_r);
				end
			else
				self:Debug(d_notice, "Skipping module " .. c_green .. v .. c_r .. " " .. c_red .. "no frame name" .. c_r);
			end
		end
		XengaInfoFrame:SetHeight(height);
	end
end

--[[
-- Sets the FrameXML scale
]]
function XengaInfo:SetFrameScale(scale)
	XengaInfoFrame:SetScale(scale);
end

--[[
-- Sets the width of the frame based on the string width
]]
function XengaInfo:CenterOneLineFrame(frame, f)
	if not frame or not frame:IsVisible() or not f then
		return;
	end;
	
	local w = 0;
	
	w = frame[f .. "Label"]:GetStringWidth() + frame[f .. "Text"]:GetStringWidth();
	
	frame:SetWidth(w);
end

--[[
-- Sets the width of the frame based on the string width of the longest row
]]
function XengaInfo:CenterTwoLineFrame(frame, f1, f2)
	if not frame or not frame:IsVisible() or not f1 or not f2 then
		return;
	end
	
	local w1, w2 = 0, 0;
	
	w1 = frame[f1 .. "Label"]:GetStringWidth() + frame[f1 .. "Text"]:GetWidth();
	w2 = frame[f2 .. "Label"]:GetStringWidth() + frame[f2 .. "Text"]:GetWidth();
	
	local new_w = 0;
	if w1 >= w2 then
		new_w = w1;
	else
		new_w = w2;
	end
	frame:SetWidth(new_w);
end

--[[
-- Takes two FontString elements and makes them the same width
]]
function XengaInfo:NormalizeStringWidths(el1, el2, padding)
	if not padding then padding = 0 end
	
	el1:SetWidth(0);
	el2:SetWidth(0);
	
	local w1, w2, new = el1:GetStringWidth(), el2:GetStringWidth(), 0;
	
	if w1 > w2 then
		new = w1;
	elseif w1 < w2 then
		new = w2;
	end
	
	new = new + padding;
	
	el2:SetWidth(new);
	el1:SetWidth(new);
end

----------------------------------------
-- END frame display functions
----------------------------------------

----------------------------------------
-- module helper functions
----------------------------------------

--[[
-- Adds a module to the disabled modules table
]]
function XengaInfo:DisableModule(m, update_frame)
	local exists = false;
	for k,v in pairs(self.sdb.disabled_modules) do
		if v == m then
			exists = true;
		end
	end
	
	if not exists then
		table.insert(self.sdb.disabled_modules, m);
	end
	
	self:RemoveModuleFrame(self.modules[m].frame);
	
	if update_frame then
		self:SetFramePositions();
	end
end

--[[
-- Removes a module from the disabled modules table
]]
function XengaInfo:EnableModule(m, update_frame)
	for k,v in pairs(self.sdb.disabled_modules) do
		if v == m then
			table.remove(self.sdb.disabled_modules, k);
			break;
		end
	end
	
	if update_frame then
		self:SetFramePositions();
	end
end
--[[
-- Returns whether the given module is active in the settings and not disabled by level restrictions
]]
function XengaInfo:IsModuleActive(m)
	return self.db.global.modules[m].show and not self:HasValue(self.sdb.disabled_modules, m);
end

--[[
-- Processes actions that need to take place at certain level thresholds
]]
function XengaInfo:ProcessLevelUp(level)
	-- disable exp
	if level == self.sdb.char.max_level then
		self:DisableModule("exp", true);
	end
	
	-- enable garrison resources
	if level == 90 then
		self:EnableModule("currency_gr", true);
	end
	
	-- enable order hall resources
	if level == 100 then
		self:EnableModule("currency_ohr", true);
	end
	
	self:UpdateCharText(level);
end

----------------------------------------
-- END module helper functions
----------------------------------------

----------------------------------------
-- module update functions
----------------------------------------

--[[
-- Triggers an update to all Update... functions
]]
function XengaInfo:UpdateAllText()
	self:UpdateCharText();
	self:UpdateCurrencyText();
	self:UpdateExpText();
	self:UpdateLatencyText();
	self:UpdateMurderPowerText();
	self:UpdateTimeText();
	self:UpdateZoneText();
end

--[[
-- Updates character level text
]]
function XengaInfo:UpdateCharText(level)
	if self.sdb.char.class == "" then
		self.sdb.char.class = UnitClass("player");
	end
	if self.sdb.char.race == "" then
		self.sdb.char.race = UnitRace("player");
	end
	if level then
		self.sdb.char.level = level;
	end
	XengaInfoTitleFrameText:SetText("Lvl " .. self.sdb.char.level .. " " .. self.sdb.char.race .. " " .. self.sdb.char.class);
end

--[[
-- Updates gold, garrison resource, and order hall resource text
]]
function XengaInfo:UpdateCurrencyText()
	-- gold
	if self:IsModuleActive("currency_gold") then
		local current_gold = GetMoney();
		local session_money = current_gold - self.sdb.currency.gold.start;
		if session_money < 0 then
			session_money = math.abs(session_money);
			XengaInfoCurrencyGoldFrame.SessionText:SetTextColor(1.0, 0, 0);
		else
			XengaInfoCurrencyGoldFrame.SessionText:SetTextColor(1.0, 1.0, 1.0);
		end
		XengaInfoCurrencyGoldFrame.CurrentText:SetText(self:GetGoldTextureString(current_gold));
		XengaInfoCurrencyGoldFrame.SessionText:SetText(self:GetGoldTextureString(session_money));
		self:NormalizeStringWidths(XengaInfoCurrencyGoldFrame.CurrentText, XengaInfoCurrencyGoldFrame.SessionText);
		self:CenterTwoLineFrame(XengaInfoCurrencyGoldFrame, "Current", "Session");
	end
	
	-- garrison resources
	if self:IsModuleActive("currency_gr") then
		local _, current_gr = GetCurrencyInfo(self.currencies.gr.id);
		local session_gr = current_gr - self.sdb.currency.gr.start;
		if session_gr < 0 then
			session_gr = math.abs(session_gr);
			XengaInfoCurrencyGrFrame.SessionText:SetTextColor(1.0, 0, 0);
		else
			XengaInfoCurrencyGrFrame.SessionText:SetTextColor(1.0, 1.0, 1.0);
		end
		XengaInfoCurrencyGrFrame.CurrentText:SetText(self:FormatNumber(current_gr, L["NUMBER_SEPARATOR"]));
		XengaInfoCurrencyGrFrame.SessionText:SetText(self:FormatNumber(session_gr, L["NUMBER_SEPARATOR"]));
		self:NormalizeStringWidths(XengaInfoCurrencyGrFrame.CurrentText, XengaInfoCurrencyGrFrame.SessionText);
		self:CenterTwoLineFrame(XengaInfoCurrencyGrFrame, "Current", "Session");
	end
	
	-- order hall resources
	if self:IsModuleActive("currency_ohr") then
		local _, current_ohr = GetCurrencyInfo(self.currencies.ohr.id);
		local session_ohr = current_ohr - self.sdb.currency.ohr.start;
		if session_ohr < 0 then
			session_ohr = math.abs(session_ohr);
			XengaInfoCurrencyOhrFrame.SessionText:SetTextColor(1.0, 0, 0);
		else
			XengaInfoCurrencyOhrFrame.SessionText:SetTextColor(1.0, 1.0, 1.0);
		end
		XengaInfoCurrencyOhrFrame.CurrentText:SetText(self:FormatNumber(current_ohr, L["NUMBER_SEPARATOR"]));
		XengaInfoCurrencyOhrFrame.SessionText:SetText(self:FormatNumber(session_ohr, L["NUMBER_SEPARATOR"]));
		self:NormalizeStringWidths(XengaInfoCurrencyOhrFrame.CurrentText, XengaInfoCurrencyOhrFrame.SessionText);
		self:CenterTwoLineFrame(XengaInfoCurrencyOhrFrame, "Current", "Session");
	end
end

--[[
-- Updates the player experience TNL and Rest text
]]
function XengaInfo:UpdateExpText()
	if self:IsModuleActive("exp") then
		-- update tnl xp
		local tnl_xp = UnitXPMax("player") - UnitXP("player");
		local tnl_xp_percent = self:Round(100-((UnitXP("player") / UnitXPMax("player")) * 100), 2);
		if tnl_xp_percent > 10 then
			tnl_xp_percent = floor(tnl_xp_percent);
		end
		XengaInfoExpFrame.TnlText:SetText(self:FormatNumber(tnl_xp, L["NUMBER_SEPARATOR"]) .. ' (' .. tnl_xp_percent .. '%)');

		--update rested xp
		local rest_xp = GetXPExhaustion();
		if rest_xp == nil then
			rest_xp = '0';
		else
			rest_xp = self:Round((rest_xp / 2), 2);
		end
		XengaInfoExpFrame.RestText:SetText(self:FormatNumber(rest_xp, L["NUMBER_SEPARATOR"]));
		self:NormalizeStringWidths(XengaInfoExpFrame.TnlText, XengaInfoExpFrame.RestText);
		self:CenterTwoLineFrame(XengaInfoExpFrame, "Tnl", "Rest");
	end
end

--[[
-- Updates the latency text
]]
function XengaInfo:UpdateLatencyText()
	if self:IsModuleActive("latency") then
		local down, up, home, world = GetNetStats();
		XengaInfoLatencyFrame.LatencyText:SetText(home .. "/" .. world);
		self:CenterOneLineFrame(XengaInfoLatencyFrame, "Latency");
	end
end

--[[
-- Updates the honor kill and item level text
]]
function XengaInfo:UpdateMurderPowerText()
	if self:IsModuleActive("murder_power") then
		local hk = GetPVPLifetimeStats();
		XengaInfoMurderPowerFrame.HKText:SetText(self:FormatNumber(hk, L["NUMBER_SEPARATOR"]));
		
		local o, e = GetAverageItemLevel();
		XengaInfoMurderPowerFrame.ItemLevelText:SetText(floor(e) .. "/" .. floor(o));
	end
end

--[[
-- Updates the session time text
]]
function XengaInfo:UpdateTimeText()
	if self:IsModuleActive("time") then
		local time = (time() - self.sdb.time_start);
		local minutes, hours = 00, 00;

		minutes = mod(floor(time / 60), 60);
		hours = mod(floor(time / 3600), 3600);

		XengaInfoTimeFrame.TimeText:SetText(format("%02d:%02d", hours, minutes));
		self:CenterOneLineFrame(XengaInfoTimeFrame, "Time");
	end
end
	
--[[
-- Updates the zone and sub-zone text
]]
function XengaInfo:UpdateZoneText()
	if self:IsModuleActive("zone") then
		local z_text, mm_text = GetRealZoneText(), GetMinimapZoneText();
		
		if z_text then
			-- set main zone
			XengaInfoZoneFrame.ZoneText:SetText(z_text);
			
			-- set sub zone
			if z_text == mm_text then
				mm_text = "-";
			else
				mm_text = '"' .. mm_text .. '"';
			end
			XengaInfoZoneFrame.SubZoneText:SetText(mm_text);
		else
			return
		end
	end
end

----------------------------------------
-- END module update functions
----------------------------------------

----------------------------------------
-- timer functions
----------------------------------------

--[[
-- Checks if the latency/time timer should be scheduled
]]
function XengaInfo:ShouldScheduleLatencyTimeTimer()
	if self.latency_time_timer == nil then
		if self:IsModuleActive("latency") or self:IsModuleActive("time") then
			return true;
		end
	end
	return false;
end

--[[
-- Checks if the latency/time timer should be stopped
]]
function XengaInfo:ShouldStopLatencyTimeTimer()
	if self.latency_time_timer == nil or self:IsModuleActive("latency") or self:IsModuleActive("time") then
		return false;
	else
		return true;
	end
end

--[[
-- Schedules the latency/time timer
]]
function XengaInfo:ScheduleLatencyTimeTimer()
	if self.latency_time_timer == nil then
		self.latency_time_timer = self:ScheduleRepeatingTimer("ProcessLatencyTimeTimer", 5);
	end
end

--[[
-- Stops the latency/time timer
]]
function XengaInfo:StopLatencyTimeTimer()
	if self.latency_time_timer ~= nil then 
		self:CancelTimer(self.latency_time_timer);
	end
end

--[[
-- Processes the latency/time timer ticks
]]
function XengaInfo:ProcessLatencyTimeTimer()
	self:UpdateLatencyText();
	self:UpdateTimeText();
end

----------------------------------------
-- END timer functions
----------------------------------------

----------------------------------------
-- utility functions
----------------------------------------

--[[
-- Dumps a table to the chat window
]]
function XengaInfo:Dump(obj, indent)
	if not indent then indent = 0 end
	local formatting;
	local spaces = "     ";
	
	for k,v in pairs(obj) do
		formatting = string.rep(spaces, indent) .. k .. " => ";
		if type(v) == "table" then
			self:Print(formatting .. "{");
			self:Dump(v, indent + 1);
			self:Print(string.rep(spaces, indent) .. "}");
		elseif type(v) == "boolean" then
			self:Print(formatting .. tostring(v));
		else
			self:Print(formatting .. v);
		end
	end
end
--[[
-- formats a number with group separators
]]
function XengaInfo:FormatNumber(number, separator)
	local formatted = FormatLargeNumber(number);
	
	if separator == nil or separator == "," then
		return formatted;
	else
		return string.gsub(formatted, ",", separator);
	end
end

--[[
-- returns a formatted string for gold
]]
function XengaInfo:GetGoldTextureString(amount)
	local gold = floor(amount / 10000);
	local silver = floor((amount - (gold * 10000)) / 100);
	local copper = mod(amount, 100);
	local gold_str, silver_str, copper_str;
	
	if self.db.global.modules.currency_gold.text_mode then
		gold_str = self:FormatNumber(gold, L["NUMBER_SEPARATOR"]) .. L["CURRENCY_ABBR_GOLD"];
		silver_str = format("%02d" .. L["CURRENCY_ABBR_SILVER"], silver);
		copper_str = format("%02d" .. L["CURRENCY_ABBR_COPPER"], copper);
	else
		gold_str =   format("%s\124TInterface\\MoneyFrame\\UI-GoldIcon:%d:%d:2:0\124t", self:FormatNumber(gold, L["NUMBER_SEPARATOR"]), 0, 0);
		silver_str = format("%02d\124TInterface\\MoneyFrame\\UI-SilverIcon:%d:%d:2:0\124t", silver, 0, 0);
		copper_str = format("%02d\124TInterface\\MoneyFrame\\UI-CopperIcon:%d:%d:2:0\124t", copper, 0, 0);
	end

	local money_str = "";
	local separator = " ";

	money_str = gold_str .. separator .. silver_str .. separator .. copper_str;

	return money_str;
end

--[[
-- Checks if the table has the given value
]]
function XengaInfo:HasValue(tab, val)
	for k,v in pairs(tab) do
		if v == val then
			return true;
		elseif type(v) == "table" then
			return self:HasValue(v, val);
		end
	end
end

--[[
-- returns a number rounded to n places
]]
function XengaInfo:Round(number, places)
	if not number then
		return 0;
	else
		local mult = math.pow(10, places or 0);
		return math.floor(number * mult + 0.5) / mult;
	end
end

----------------------------------------
-- END utility functions
----------------------------------------

----------------------------------------
-- Debug functions
----------------------------------------

--[[
-- prints an error message
]]
function XengaInfo:Debug(level, message, ...)
	if self.db.global.debug_level then
		if self.db.global.debug_level ~= d_off and level <= self.db.global.debug_level then
			self:Printf(WrapTextInColorCode(debug_levels[level].name, "ff" .. debug_levels[level].color) .. ": " .. message, ...)
		end
	end
end

--[[
-- gets current debug level
]]
function XengaInfo:GetDebugLevel()
	return self.db.global.debug_level;
end

--[[
-- sets current debug level
]]
function XengaInfo:SetDebugLevel(level)
	if (level == nil) then
		level = d_off;
	end
	if debug_levels[level] then
		local name = debug_levels[level].name;
		self:Printf(L["MSG_DEBUG_LEVEL_SET"], name);
		self.db.global.debug_level = level;
	else
		self:Printf(L["MSG_DEBUG_LEVEL_INVALID"], "ff" .. debug_levels[d_warn].color, level);
	end
end

----------------------------------------
-- END Debug functions
----------------------------------------
